<?php
namespace nur\sery\wip\schema\_assoc;

use nulib\cl;
use nulib\ref\schema\ref_analyze;
use nulib\ValueException;
use nur\sery\wip\schema\_scalar\ScalarResult;
use nur\sery\wip\schema\_scalar\ScalarWrapper;
use nur\sery\wip\schema\input\Input;
use nur\sery\wip\schema\Result;
use nur\sery\wip\schema\types\IType;
use nur\sery\wip\schema\Wrapper;
use nur\sery\wip\schema\WrapperContext;

class AssocWrapper extends Wrapper {
  function __construct(AssocSchema $schema, &$value=null, $valueKey=null, ?array $params=null) {
    $keys = $schema->getKeys();
    $keyParams = cl::merge($params, [
      "throw" => false,
    ]);
    $keyWrappers = [];
    foreach ($keys as $key) {
      $value = null;
      $keyWrappers[$key] = $schema->getSchema($key)->getWrapper($value, null, $keyParams);
    }
    $this->context = $context = new AssocWrapperContext($schema, null, null, $params);
    $arrayParams = cl::merge($params, [
      "throw" => false,
    ]);
    $context->arrayWrapper = new ScalarWrapper($schema, $dummy, null, $arrayParams, $context);
    $context->keys = $keys;
    $context->keyWrappers = $keyWrappers;

    if ($value !== null) {
      # n'initialiser que si $value n'est pas null
      $this->reset($value, $valueKey);
    }
  }

  /** @var AssocWrapperContext  */
  protected WrapperContext $context;

  protected function resetContext($resetSelectedKey): void {
    parent::resetContext($resetSelectedKey);
    $context = $this->context;
    $context->arrayWrapper->getResult()->reset();
    foreach ($context->keyWrappers as $wrapper) {
      $wrapper->getResult()->reset();
    }
  }

  function reset(&$value, $valueKey=null, ?array $params=null): Wrapper {
    $context = $this->context;
    if ($value instanceof Input) $input = $value;
    else $input = $this->newInput($value);
    $context->input = $input;
    $context->valueKey = $valueKey;
    foreach ($context->keyWrappers as $key => $keyWrapper) {
      $keyInput = $input->addKey($valueKey);
      $keyWrapper->reset($keyInput, $key, ["analyze" => false]);
    }
    $this->afterModify($params, true);
    return $this;
  }

  function getKeys(): array {
    return $this->context->keys;
  }

  protected function _getWrapper($key): Wrapper {
    $context = $this->context;
    if ($key === null) return $context->arrayWrapper;
    $wrapper = $context->keyWrappers[$key] ?? null;
    if ($wrapper === null) throw ValueException::invalid_key($key);
    return $wrapper;
  }

  /** @param string|int|null $key */
  function select($key=null): Wrapper {
    $wrapper = $this->_getWrapper($key);
    $this->context->selectedKey = $key;
    return $wrapper;
  }

  /**
   * @param AssocWrapperContext $context
   * @param AssocWrapper $wrapper
   */
  static function _analyze(WrapperContext $context, Wrapper $wrapper, ?array $params): int {
    if ($params["ensure_array"] ?? $context->ensureArray) {
      $valueKey = $context->valueKey;
      $array = $context->input->get($valueKey);
      if ($array === null) $context->input->set([], $valueKey);
    }

    if ($params["ensure_assoc"] ?? $context->ensureAssoc) {
      $context->input->ensureAssoc($context->schema->getKeys());
    }

    $what = ScalarWrapper::_analyze($context, $wrapper, $params);
    /** @var ScalarResult $result */
    $result = $context->result;
    if (!$result->valid) return $what;

    foreach ($context->keyWrappers as $keyWrapper) {
      $keyWrapper->analyze($params);
      if (!$keyWrapper->isValid()) {
        #XXX distinguer MISSING, UNAVAILABLE, NULL et !VALID
        $what = ref_analyze::INVALID;
        $result->addInvalidMessage($keyWrapper);
      }
    }

    #XXX supprimer les clés "missing" ou "unavailable" sauf si $ensureKeys
    return $what;
  }

  /**
   * @param AssocWrapperContext $context
   * @param AssocWrapper $wrapper
   */
  static function _normalize(WrapperContext $context, Wrapper $wrapper, ?array $params): bool {
    $ensureKeys = $params["ensure_keys"] ?? $context->ensureKeys;
    $ensureOrder = $params["ensure_order"] ?? $context->ensureOrder;
    if ($ensureKeys || $ensureOrder) {
      $schema = $context->schema;
      $keys = $schema->getKeys();
      if ($ensureKeys) {
        $defaults = [];
        foreach ($keys as $key) {
          $default = $schema->getSchema($key)->default;
          if ($default === null) {
            $default = $wrapper->getType($key)->getNullValue();
          }
          $defaults[$key] = $default;
        }
      }
      if ($ensureKeys) $context->input->ensureKeys($defaults, $params);
      if ($ensureOrder) $context->input->ensureOrder($keys, $params);
    }

    $modified = ScalarWrapper::_normalize($context, $wrapper, $params);
    foreach ($context->keyWrappers as $keyWrapper) {
      if ($keyWrapper->normalize($params)) $modified = true;
    }
    return $modified;
  }

  function getResult($key=false): Result {
    if ($key === false) $key = $this->context->selectedKey;
    return $this->_getWrapper($key)->getResult();
  }

  function getType($key=false): IType {
    if ($key === false) $key = $this->context->selectedKey;
    return $this->_getWrapper($key)->getType();
  }

  function get($default=null, $key=false) {
    $context = $this->context;
    if (!$context->arrayWrapper->isAvailable()) return $default;
    if ($key === false) $key = $context->selectedKey;
    return $this->_getWrapper($key)->get($default);
  }

  function set($value, ?array $params=null, $key=false): Wrapper {
    $context = $this->context;
    if ($key === false) $key = $context->selectedKey;
    $this->_getWrapper($key)->set($value);
    return $this;
  }

  function unset(?array $params=null, $key=false): Wrapper {
    $context = $this->context;
    if ($key === false) $key = $context->selectedKey;
    $this->_getWrapper($key)->unset();
    return $this;
  }

  function format($format=null, $key=false): string {
    $context = $this->context;
    if ($key === false) $key = $context->selectedKey;
    return $this->_getWrapper($key)->format($format);
  }
}
